% This file has been written for trying to re-understand how the database
% and the scripts are working.
% Run it to change current folder

error('Do not run this script as a standalon script, use sections!');

%% Step 1: solver V4
%
% the V4 is actually more complex and more difficult to use with respect to
% the V3. It has been written to solve many issues, e.g., the body diode is
% modeled, different zone succession are considered, etc.
%
% INPUT:
% iinv0, irec0, vka0: initial point [note: theta=0 the MOS has been turned
%                     ON, in theta=theta2 the MOS is turned OFF, and in
%                     theta=2*pi it is turned ON again. ZVS and ZVDS has to
%                     be checked at 2*pi
% theta2: MOS turn OFF time [note: Duty Cycle is theta2/2/pi]
% qI, qR, qM, kI, kR: dimensionless values
% vdON, vbON: rectifying diode and body diode voltage drop
%             [note: normalized with respect to Vinv and Vrec,
%             respectively]
% QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT: lossy parameters
%
% OUTPUT:
% [V, fZ, thetas, zzzz] = solver4(...)
% V: vector for minimization. 
%    V(1:4) is [iinv(2*pi); irec(2*pi); vds(2*pi); vka(2*pi)]
%    V(5) is 'lastboff' the last time when body diode was off
%    V(6) is 'iirec' integrate(irec(theta), 0, 2*pi)
% fZ: functions to plot the waveforms
% thetas: array for the angles of changing zones
% zzzz: string with the different zones

clear all; % to be sure that all variables are set in this section
close all;

kI = 0.5;
kR = 0.5;
qI = 1.3773;
qR = 1.3773;
qM = 0.4653;

theta2 = 0.5*2*pi;

iinv0 = 0;
irec0 = -1.8628;
vka0 = 3.1095;

vdON = 0.;
vbON = 0.;
QLi = inf;
QLr = inf;
QM = inf;
QCinv = inf;
QCrec = inf; 
gIN = inf;
gdsON = inf;
gbON = inf;
gCM = inf;
gdON = inf;
gOUT = inf;


[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);


%% Step 2: Minimization (ZVS + ZVDS / desired current value not zero)
%
% 1. given iinv0, irec0, vka0, continuity has to be ensured:
%      V([1;2;4]) = [iinv0; irec0; vka0]
% 2. ZVS has to be ensured
%      V(3) = 0
% 3. ZVDS has to be ensured / desired current level to be esured
%      V(1) = iinv0 && (iinv0=0 || iinv0 = the desired current level)
% 4. We do not want that the body diode turns on
%      V(5) = 2*pi
% 5. We want unitary output power
%      V(6) = -2*pi
%
% So: set iinv0 to desired level && V = [iinv0; irec0; 0; vka0; 2*pi; -2*pi]
% optimization variables (kI and kR fixed): qI, qR, qM, irec0, vka0

clear all; % to be sure that all variables are set in this section
close all;

kI = 0.5;
kR = 0.5;
qI = 1.3773;
qR = 1.3773;
qM = 0.4653;

theta2 = 0.5*2*pi;

iinv0 = -0.25; % Example: not ZVDS, but residual current to 0.25
irec0 = -1.8628;
vka0 = 3.1095;

vdON = 0.;
vbON = 0.;
QLi = inf;
QLr = inf;
QM = inf;
QCinv = inf;
QCrec = inf; 
gIN = inf;
gdsON = inf;
gbON = inf;
gCM = inf;
gdON = inf;
gOUT = inf;

x0 = [qI, qR, qM, irec0, vka0];
fToMinimize = @(x5) norm(...
    solver4b (iinv0, x5(4), x5(5), theta2, x5(1), x5(2), x5(3), kI, kR, ...
        vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gbON, gdsON, gCM, gdON, gOUT) ...
    - [iinv0; x5(4); 0; x5(5); 2*pi; -2*pi]);

tic;
[xoptimal, residual] = fminsearch(fToMinimize, x0 ,optimset('TolFun',1e-9,'MaxFunEvals',2500));       
toc;

disp(xoptimal);
disp(residual);

irec0 = xoptimal(4);
vka0 = xoptimal(5);
qI = xoptimal(1);
qR = xoptimal(2);
qM = xoptimal(3);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% Step 3: Minimization ZVS only (any current is ok)
%
% 1. given iinv0, irec0, vka0, continuity has to be ensured:
%      V([1;2;4]) = [iinv0; irec0; vka0]
% 2. ZVS has to be ensured
%      V(3) = 0
% 4. We do not want that the body diode turns on
%      V(5) = 2*pi
% 5. We want unitary output power
%      V(6) = -2*pi
%
% So: V = [iinv0; irec0; 0; vka0; 2*pi; -2*pi]
% optimization variables (kI and kR fixed): qI, qR, qM, irec0, iinv0, vka0

clear all; % to be sure that all variables are set in this section
close all;

kI = 0.5;
kR = 0.5;
qI = 1.45;
qR = 1.45;
qM = 0.4653;

theta2 = 0.5*2*pi;

iinv0 = -0.25;
irec0 = -1.8628;
vka0 = 3.1095;

vdON = 0.;
vbON = 0.;
QLi = inf;
QLr = inf;
QM = inf;
QCinv = inf;
QCrec = inf; 
gIN = inf;
gdsON = inf;
gbON = inf;
gCM = inf;
gdON = inf;
gOUT = inf;

% state space has 6 variables 
x0 = [qI, qR, qM, iinv0, irec0, vka0];
fToMinimize = @(x6) norm(...
    solver4b (x6(4), x6(5), x6(6), theta2, x6(1), x6(2), x6(3), kI, kR, ...
        vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT) ...
    - [x6(4); x6(5); 0; x6(6); 2*pi; -2*pi]);

tic;
[xoptimal, residual] = fminsearch(fToMinimize, x0 ,optimset('TolFun',1e-9,'MaxFunEvals',2500));       
toc;

disp(xoptimal);
disp(residual);

iinv0 = xoptimal(4);
irec0 = xoptimal(5);
vka0 = xoptimal(6);
qI = xoptimal(1);
qR = xoptimal(2);
qM = xoptimal(3);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% step 3a - mutual minimization of residual and iinv0 current
%
% solver4 is written as a function:
% (qI, qR, qM, iinv0, irec0, vka0) -> (iinv(2pi-), irec(2pi-), vds(2pi-), vka(2pi-), lastboff, iirec)
% X6 -> V, starting from X0
% we need to set to zero (at least 1e-3) norm(V - [iinv0; irec0; 0; vka0; 2*pi; -2*pi])
% and, at the same time, minimize abs(iinv0)
%
% abs(iinv0) ranges about from 0 to 100 - so the idea is to minimize
%    abs(iinv0) + f(norm(V-[iinv0; irec0; 0; vka0; 2*pi; -2*pi]))
%    abs(X(4)) + f(norm(V-[X(4); X(5); 0; X(6); 2*pi; -2*pi]))
%
%    if norm(...) < 1e-4 -> f(norm(...)) negligible
%    if norm(...) > 1e-3 -> f(norm(...)) very large
%    f(1e-4) = 1e-4 [target value for abs(iinv0)=0]
%    f(1e-3) = 100  [maximum value of abs(iinv0)]
%    f(...) = (1e3*...).^6*1e2
%    test = (1e3 * [1e-5 1e-4 1e-3 1e-2]).^6 * 1e2; -> [1e-10 1e-4 1e2 1e8]
%

clear all; % to be sure that all variables are set in this section
close all;

kI = .4;
kR = .4;
qI = 2.5;
qR = 1.5;
qM = 0.32;

theta2 = 0.7*2*pi;

iinv0 = -2.55;
irec0 = -1.07;
vka0 = 3.88;

vdON = 0.;
vbON = 0.;
QLi = inf;
QLr = inf;
QM = inf;
QCinv = inf;
QCrec = inf; 
gIN = inf;
gdsON = inf;
gbON = inf;
gCM = inf;
gdON = inf;
gOUT = inf;

% state space has 6 variables 
x0 = [qI, qR, qM, iinv0, irec0, vka0];
finv0res4 = @(iinv0, res4) abs(iinv0) + 1e2*(1e3*res4)^6;
fToMinimize = @(x6) finv0res4(x6(4), norm(...
    solver4b (x6(4), x6(5), x6(6), theta2, x6(1), x6(2), x6(3), kI, kR, ...
        vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT) ...
    - [x6(4); x6(5); 0; x6(6); 2*pi; -2*pi]));
tic;
[xoptimal, residual] = fminsearch(fToMinimize, x0 ,optimset('TolFun',1e-9,'MaxFunEvals',2500));       
toc;

disp(xoptimal);
disp(residual);

iinv0 = xoptimal(4);
irec0 = xoptimal(5);
vka0 = xoptimal(6);
qI = xoptimal(1);
qR = xoptimal(2);
qM = xoptimal(3);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    vdON, vbON, QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gbON, gCM, gdON, gOUT);

outlongline = line_computeresiduals_v4b([kI, kR, qI, qR, qM, iinv0, irec0, vka0], theta2);


disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% step 4 - The database
%
% DCoptimal-v4h1-std-05.cs
%     ^        ^  ^   ^
%     ^        ^  ^   ^--- Duty cicle (theta2 = 2*pi*duty cycle)
%     ^        ^  ^------- coupling, std or inv
%     ^        ^---------- harmonic: first, secondo, or more (to be manually detected...)
%     ^------------------- optimal (ZVS + ZVDS) or relaxed (ZVS + minimum dv/dt)
% 
% All databases are assuming an ideal system (vxON=0, Qxxx=inf, gxxx=inf)

clear all;
close all;

headers = {'kI', 'kR', 'qI', 'qR', 'qM', 'iinv0', 'irec0', 'vka0', ... % 1 to 8 [8]
           'iinv2pi[ZVDS]', 'irec2pi', 'vds2pi[ZVS]', 'vka2pi', 'lasttbon', 'iiF', ... % 9 to 14 [6] 
           'theta2', 'deprecated', ... % 15 to 16 [2] 
           'resoptimal', 'resrelaxed', 'lastZVDSreduction', ... % 17 to 19 [3]
           'spiceiinv', 'spiceirec', 'spicevds', 'spicevka', 'spiceall', ... % 20 to 24 [5]
           'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved'}; % 25 to 32 [8]
%  
% 'kI', 'kR', 'qI', 'qR', 'qM', 'iinv0', 'irec0', 'vka0' are self-explaining
% 'iinv2pi[ZVDS]', 'irec2pi', 'vds2pi[ZVS]', 'vka2pi', 'lasttbon', 'iiF' still self-explaining
%   iinv(2*pi)=ZVDS is the value of iinv at the MOS turn on instant (should be zero, or minimized in module in relaxed mode)
%   vds(2*pi)=ZVS is the value of vCinv/vds at the MOS turn on instant (should be zero) 
%   lasttbon the last time when body diode turned on (2*pi not turned on)
%   iiF is the value of Integrate[irec[t],t,0,2*pi] (should be -2*pi)
% theta2 is the normalized time when the MOS turn OFF
% resoptimal is the residual value using the optimal design (ZVS + ZVDS)
% resrelaxed is the residual value using the relaxed design (ZVS only + optional current minimization)
% lastZVDSreduction is the last reduction for which convegence was not reached [0.1, 0.01, or 0.001]
% spiceiinv, spiceirec, spicevds, spicevka are the (normalized) integral of the differences between spice and matlab waveforms
% unused (or uncomputed) fields are set to -1.

FILEoptimal = 'DCoptimal-v4h1-std-05.csv';
FILErelaxed = 'DCrelaxed-v4h1-std-05.csv';

DBoptimal = dlmread(FILEoptimal,',',1,0);
DBrelaxed = dlmread(FILErelaxed,',',1,0);

theline = DBoptimal(45,:); % disp(theline(17)) % check the residual 

kI = theline(1);
kR = theline(2);
qI = theline(3);
qR = theline(4);
qM = theline(5);

theta2 = theline(15);

iinv0 = theline(6);
irec0 = theline(7);
vka0 = theline(8);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    0, 0, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        figure;
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

theline = DBrelaxed(45,:); % disp(theline(18)) % check the residual 

kI = theline(1);
kR = theline(2);
qI = theline(3);
qR = theline(4);
qM = theline(5);

theta2 = theline(15);

iinv0 = theline(6);
irec0 = theline(7);
vka0 = theline(8);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    0, 0, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        figure;
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% Step 5: database function: line_extract_nearest_v4b.m
%
% Get the line nearest to the asked one. If existing, gives the line; if not 
% existing, create the initial point for refinement
% note: short line is [1:8] (only minimal information)
%       long line is 1:32 (everything)
%
% usage: [line, distance] = line_extract_nearest_v4b (DB, [kI, kR])
%        line = line_extract_nearest_v4b (DB, [kI, kR])
% DB is the BD to be used, [kI, kR] the point, distance the distance from
% the asked point to the one found, line the (long) line

clear all;
close all;

%FILEoptimal = 'DCoptimal-v4h1-std-05.csv';
%FILEoptimal = 'DCoptimal-v4h1-std-03.csv';
FILEoptimal = 'DCrelaxed-v4h1-std-03.csv';
DBoptimal = dlmread(FILEoptimal,',',1,0);

%[theline, distance] = line_extract_nearest_v4b(DBoptimal, [0.4, 0.4]);
%[theline, distance] = line_extract_nearest_v4b(DBoptimal, [0.775, 0.9]);
[theline, distance] = line_extract_nearest_v4b(DBoptimal, [0.2, 0.2]);

disp(distance);

kI = theline(1);
kR = theline(2);
qI = theline(3);
qR = theline(4);
qM = theline(5);

theta2 = theline(15);

iinv0 = theline(6);
irec0 = theline(7);
vka0 = theline(8);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    0, 0, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        figure;
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% Step 6: database function: line_refine_v4bshort.m
%
% Try to refine the input (short) line
%
% usage: [out, residual] = line_refine_v4bshort(in, target, theta2[, tobeavoid])
%        out = line_refine_v4bshort(in, target, theta2[, tobeavoid])
%
%  out is the out line (short line), residual the minimization residual
%  in the starting point (short line)
%  target the optimization target: [ZVDS; ZVS; lastbon; iif] 
%         (i.e., setting ZVDS to the desired level, ZVS to the desired level)
%         for ZVS + ZVDS => [0; 0; 2*pi; -2*pi] 
%         for ZVS only, any ZVDS => [NaN; 0; 2*pi; -2*pi] 
%  theta2 self-explains
%  tobeavoid: matrix of short-line to be avoided (to be implemented)

clear all;
close all;

FILEoptimal = 'DCoptimal-v4h1-std-05.csv';
DBoptimal = dlmread(FILEoptimal,',',1,0);

theta2 = 0.5*2*pi;

[inline, distance] = line_extract_nearest_v4b(DBoptimal, [0.4, 0.4]);
inshortline = inline(1:8);

%target = [NaN; 0; 2*pi; -2*pi];
target = [0; 0; 2*pi; -2*pi];
%target = [-0.1; 0; 2*pi; -2*pi];

[outshortline, residual] = line_refine_v4bshort(inshortline, target, theta2);
disp(residual);

kI = outshortline(1);
kR = outshortline(2);
qI = outshortline(3);
qR = outshortline(4);
qM = outshortline(5);

iinv0 = outshortline(6);
irec0 = outshortline(7);
vka0 = outshortline(8);

[V, fZ, thetas, zzzz] = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    0, 0, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf, inf);
 
disp (V);

        T = 2*pi*(0:255)'/255;
        Y = zeros(256,4);
        for ii=1:256
            Y(ii,:) = fZ(T(ii));
        end
        figure;
        plot([T; T+2*pi], [Y; Y])
        
disp (thetas);

disp (zzzz);

%% Step 7: database functions: line_computeresiduals_v4b.m, line_add_or_replace_vall.m & line_remove_vall.m
%
% Aux functions
%
% usage: longline = line_computeresiduals_v4b.m(shortline, theta2)
%        compute the longline from the short line
%
% usage: database = line_add_or_replace_vall.m(database, longline)
%        add or replace the line in the database
%
% usage: database = line_remove_vall.m(database, longline)
%        remove the line in the database

clear all;
close all;

FILEoptimal = 'DCoptimal-v4h1-std-05.csv';
DBoptimal = dlmread(FILEoptimal,',',1,0);

theta2 = 0.5*2*pi;

[inline, distance] = line_extract_nearest_v4b(DBoptimal, [0.4, 0.4]);
inshortline = inline(1:8);

%target = [NaN; 0; 2*pi; -2*pi];
target = [0; 0; 2*pi; -2*pi];
%target = [-0.1; 0; 2*pi; -2*pi];

[outshortline, residual] = line_refine_v4bshort(inshortline, target, theta2);
disp(residual);

outlongline = line_computeresiduals_v4b(outshortline, theta2);
DBoptimal = line_remove_vall(DBoptimal,outlongline(1:2)); % optional
DBoptimal = line_add_or_replace_vall(DBoptimal,outlongline);

%% Step 8: update the whole database 03 (takes between between 15 and 30 secs depending on size)

clear all;
close all;

headers = {'kI', 'kR', 'qI', 'qR', 'qM', 'iinv0', 'irec0', 'vka0', ... % 1 to 8 [8]
           'iinv2pi[ZVDS]', 'irec2pi', 'vds2pi[ZVS]', 'vka2pi', 'lasttbon', 'iiF', ... % 9 to 14 [6] 
           'theta2', 'deprecated', ... % 15 to 16 [2] 
           'resoptimal', 'resrelaxed', 'lastZVDSreduction', ... % 17 to 19 [3]
           'spiceiinv', 'spiceirec', 'spicevds', 'spicevka', 'spiceall', ... % 20 to 24 [5]
           'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved'}; % 25 to 32 [8]

FILEoptimal = 'DCoptimal-v4h1-std-03.csv';
%FILEoptimal = 'DCoptimal-v4h1-inv-03.csv';
%FILEoptimal = 'DCrelaxed-v4h1-std-03.csv';
%FILEoptimal = 'DCrelaxed-v4h1-inv-03.csv';
DBoptimal = dlmread(FILEoptimal,',',1,0);
theta2 = 0.3*2*pi;

tic;
for i = 1:size(DBoptimal,1)
    
    line = DBoptimal(i,:);
    lineshort = line(1:8);
    outlongline = line_computeresiduals_v4b(lineshort, theta2);
    DBoptimal = line_add_or_replace_vall(DBoptimal,outlongline); 

end
toc;

csvwrite_with_headers_12g(FILEoptimal, DBoptimal, headers); disp('database updated.');

%% Step 9: example of expanding database 03 (optimal)

clear all;
close all;

headers = {'kI', 'kR', 'qI', 'qR', 'qM', 'iinv0', 'irec0', 'vka0', ... % 1 to 8 [8]
           'iinv2pi[ZVDS]', 'irec2pi', 'vds2pi[ZVS]', 'vka2pi', 'lasttbon', 'iiF', ... % 9 to 14 [6] 
           'theta2', 'deprecated', ... % 15 to 16 [2] 
           'resoptimal', 'resrelaxed', 'lastZVDSreduction', ... % 17 to 19 [3]
           'spiceiinv', 'spiceirec', 'spicevds', 'spicevka', 'spiceall', ... % 20 to 24 [5]
           'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved'}; % 25 to 32 [8]

FILEoptimal = 'DCoptimal-v4h1-std-03.csv';
%FILEoptimal = 'DCoptimal-v4h1-inv-03.csv';
%FILEoptimal = 'DCrelaxed-v4h1-std-03.csv';
%FILEoptimal = 'DCrelaxed-v4h1-inv-03.csv';
DBoptimal = dlmread(FILEoptimal,',',1,0);
theta2 = 0.3*2*pi;

kIlist = 0.8-0.025*(0:31);
%kIlist=[0.2 0.175 0.15 0.125];
kR = 1.025;
%kI = 0.2;
%kRlist = 0.4-0.025*[0:15];

for kI = kIlist
%for kR = kRlist
    
    [line, distance] = line_extract_nearest_v4b(DBoptimal, [kI, kR], [1e-3, 1e-3]);
    shortline = line(1:8);
    prevres5 = line(17);
    target = [0; 0; 2*pi; -2*pi];
    [outshortline, residual] = line_refine_v4bshort(shortline, target, theta2);
    outlongline = line_computeresiduals_v4b(outshortline, theta2); 
    res5 = outlongline(17);
    disp([kI, kR, prevres5, res5]);
    % ATTENTION: residual may be different from res5 or res4
    % Why? Only numerical approximations or is there an error somewhere?

    if (res5 < 1e-3) & (res5 < prevres5) % residual better than existing

        disp('updated');
        %DBoptimal = line_remove_vall(DBoptimal,outlongline(1:2)); % optional
        DBoptimal = line_add_or_replace_vall(DBoptimal,outlongline); 

    end
    
end

csvwrite_with_headers_12g(FILEoptimal, DBoptimal, headers); disp('database updated.');

%% Step 10: example of expanding database 03 (relaxed)

clear all;
close all;

finv0res4 = @(iinv0, res4) abs(iinv0) + 1e2*(1e3*res4)^6;

headers = {'kI', 'kR', 'qI', 'qR', 'qM', 'iinv0', 'irec0', 'vka0', ... % 1 to 8 [8]
           'iinv2pi[ZVDS]', 'irec2pi', 'vds2pi[ZVS]', 'vka2pi', 'lasttbon', 'iiF', ... % 9 to 14 [6] 
           'theta2', 'deprecated', ... % 15 to 16 [2] 
           'resoptimal', 'resrelaxed', 'lastZVDSreduction', ... % 17 to 19 [3]
           'spiceiinv', 'spiceirec', 'spicevds', 'spicevka', 'spiceall', ... % 20 to 24 [5]
           'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved', 'reserved'}; % 25 to 32 [8]

%FILEoptimal = 'DCoptimal-v4h1-std-03.csv';
FILErelaxed = 'DCrelaxed-v4h1-std-03.csv';
%DBoptimal = dlmread(FILEoptimal,',',1,0);
DBrelaxed = dlmread(FILErelaxed,',',1,0);
theta2 = 0.3*2*pi;


kI = 0.2;
kRlist = 0.25-0.025*[0:4];

for kR = kRlist
    
    [line, distance] = line_extract_nearest_v4b(DBrelaxed, [kI, kR]);
    shortline = line(1:8);
%    prevres4 = line(18);
    previnv0res4 = finv0res4(line(6),line(18));
    target = [NaN; 0; 2*pi; -2*pi];
    [outshortline, residual] = line_refine_v4bshort(shortline, target, theta2);
    outlongline = line_computeresiduals_v4b(outshortline, theta2); 
    res5 = outlongline(17); % if optimized, res5 is the absolute value of the current
    res4 = outlongline(18);
    inv0res4 = finv0res4(outlongline(6),outlongline(18));
    disp([kI, kR, res5, res4]);
    % ATTENTION: residual may be different from res5 or res4
    % Why? Only numerical approximations or is there an error somewhere?

    if (res4 < 1e-3) & (inv0res4 < previnv0res4) % poit is good and better than existing

        disp('updated');
        %DBoptimal = line_remove_vall(DBoptimal,outlongline(1:2)); % optional
        DBrelaxed = line_add_or_replace_vall(DBrelaxed,outlongline); 

    end
    
end

csvwrite_with_headers_12g(FILErelaxed, DBrelaxed, headers); disp('database updated.');

