function varargout = solver4b (iinv0, irec0, vka0, theta2, qI, qR, qM, kI, kR, ...
    vdON, vbON, QI, QR, QM, QCinv, QCrec, gINV, gdsON, gbON, gCM, gdON, gREC)
% the NEW solver
% both direct coupling and reverse coupling start at theta=0 when the MOS
% is turned ON. The mos is turned OFF at theta2. Need to check continuity
% of iinv(theta), irec(theta), vka(theta)
%
% the V4 is actually more complex and more difficult to use with respect to
% the V3. It has been written to solev many issues, e.g., the body diode is
% modeled, different zone succession are considered, etc.

% USAGE:
% Input parameters:
% iinv0, irec0, vka0: initial point [note: theta=0 the MOS has been turned
%                     ON, in theta=theta2 the MOS is turned OFF, and in
%                     theta=2*pi it is turned ON again. ZVS and ZVDS has to
%                     be checked at 2*pi
% theta2: MOS turn OFF time [note: Duty Cycle is theta2/2/pi]
% qI, qR, qM, kI, kR: dimensionless values
% vdON, vbON: rectifying diode and body diode voltage drop
%             [note: normalized with respect to Vinv and Vrec,
%             respectively]
% QLi, QLr, QM, QCinv, QCrec, gIN, gdsON, gCM, gdON, gOUT: lossy parameters
%
% Output parameters : [V, fZ, thetas, zzzz] = solver4(...)
% V: vector for minimization. 
%    V(1:4) is [iinv(2*pi); irec(2*pi); vds(2*pi); vka(2*pi)]
%    V(5) is 'lastboff' the last time when body diode was off
%    V(6) is 'iirec' integrate(irec(theta), 0, 2*pi)
% fZ: functions to plot the waveforms
% thetas: array for the angles of changing zones
% zzzz: string with the different zones

% CHANGELIST:
% 4/11/2019: removal of short zones around 2*pi if extension < 2pi/100 
%            (avoid a lot of useless micro-zones around 2pi)
% 27/11/2019: Check also for qI!=0 && qR!=0 && kI!=0 && kR!=0 && kI kR > 1
%            (funny solutions for qR=0 [Crec=Inf], with Vka=1 constant)
% 13/3/2020: Added the gbON parameter, correction to Z3a and Z4a zones
% 10/4/2020: create more flexible diode-on condition
% 30/4/2020: new smart condition on lastboff (that is now firstbon)
%
% TODO:

    % do not perform analysis if:
    % - kI*kR >= 1 (no solution)
    % - kI, kR = 0 (no solution)
    % - qI, qR \approx 0 -> a solution can be found assuming infinite Cinv
    %                       and Crec (vDS and vKA constant, unrelated to
    %                       iinv and irec)
    %   *** iinv = 0, irec=-1, vds = 0, vka > vdon is a constant solution
    %       if qI->0, qR->0. 
    %       Assuming i_rec = 1 -> vrec(2*pi) - vrec0 = qR integrate(irec, 0, 2*pi) 
    %       (rectifying diode is on always)
    %       so vrec(2*pi) - vrec0 = 2*pi*qR
    %       If I want to impose vrec - vrec0 > 1e-4 (convergence failure) 
    %       then it has to be qR < 1e-4 / (2*pi)
    if abs(kI)<1e-5 || abs(kR)<1e-5 || kI*kR >= 1 || ... 
            abs(qI)<1e-3 || abs(qR)<1e-3 || abs(qI)>1e3 || abs(qR)>1e3 || abs(qM)>1e3
        % here no solution -- to be handled as it can be reached during minimization             
        [varargout{1:nargout}] = reterror(iinv0, irec0, vka0);
        return
    end



    AAA1 = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(qI/QCinv + 1/gINV)), ... AAA1(1,1)
             kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(1/gdON + 1/gREC)), ... AAA1(1,2)
             -1/(1 - kI*kR) * kI / qM; ... % AAA1(1,3)
             kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(qI/QCinv + 1/gINV)), ... AAA1(2,1)
             -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(1/gdON + 1/gREC)), ... AAA1(2,2)
             kR/(1 - kI*kR) * kI / qM; ... AAA1(2,3)
             qI, 0, 0]; % AAA1(3,:)

    AAA2 = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(qI/QCinv + 1/gINV)), ... AAA2(1,1)
             kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA2(1,2)
             -1/(1 - kI*kR) * kI / qM, ... AAA2(1,3)
             kI/(1 - kI*kR) * kR / qM; ... AAA2(1,4)
             kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(qI/QCinv + 1/gINV)), ... AAA2(2,1)
             -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA2(2,2)
             kR/(1 - kI*kR) * kI / qM, ... AAA2(2,3)
             -1/(1 - kI*kR) * kR / qM; ... AAA2(2,4)
             qI, 0, 0, 0; ... AAA2(3,:)
             0, qR, 0, 0]; % AAA2(4,:)

    AAA3 = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(1/gdsON + 1/gINV)), ... AAA3(1,1)
             kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA3(1,2)
             kI/(1 - kI*kR) * kR / qM; ... AAA3(1,3)
             kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(1/gdsON + 1/gINV)), ... AAA3(2,1)
             -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA3(2,2)
             -1/(1 - kI*kR) * kR / qM; ... AAA3(2,3)
             0, qR, 0]; % AAA3(3,:)
         
    AAA4 = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(1/gdsON + 1/gINV)), ... AAA4(1,1)
             kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(1/gdON + 1/gREC)); ... AAA4(1,2)
             kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(1/gdsON + 1/gINV)), ... AAA1(2,1)
             -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(1/gdON + 1/gREC))]; % AAA1(2,2)

    AAA3a = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(1/gbON + 1/gINV)), ... AAA3(1,1)
              kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA3(1,2)
              kI/(1 - kI*kR) * kR / qM; ... AAA3(1,3)
              kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(1/gbON + 1/gINV)), ... AAA3(2,1)
              -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(qR/QCrec + 1/gREC)), ... AAA3(2,2)
              -1/(1 - kI*kR) * kR / qM; ... AAA3(2,3)
              0, qR, 0]; % AAA3(3,:)
         
    AAA4a = [ -1/(1 - kI*kR) * ((1-kI)/QI + kI*(1-kR)/QM + kI*(1-kR)/qM/gCM + kI/qM*(1/gbON + 1/gINV)), ... AAA4(1,1)
              kI/(1 - kI*kR) * ((1-kR)/QR - (1-kR)/QM - (1-kR)/qM/gCM + kR/qM*(1/gdON + 1/gREC)); ... AAA4(1,2)
              kR/(1 - kI*kR) * ((1-kI)/QI - (1-kI)/QM - (1-kI)/qM/gCM + kI/qM*(1/gbON + 1/gINV)), ... AAA1(2,1)
              -1/(1 - kI*kR) * ((1-kR)/QR + kR*(1-kI)/QM + kR*(1-kI)/qM/gCM + kR/qM*(1/gdON + 1/gREC))]; % AAA1(2,2)

    BBB1 = [ 1/(1 - kI*kR) * kI/qM * (1 - kR - kR * vdON); ...
             1/(1 - kI*kR) * kR/qM * (1 - kI + vdON); 0];

    BBB2 = [ 1/(1 - kI*kR) * kI/qM * (1 - kR); ... 
             1/(1 - kI*kR) * kR/qM * (1 - kI); 0; 0];

    BBB3 = [ 1/(1 - kI*kR) * kI/qM * (1 - kR); ...
             1/(1 - kI*kR) * kR/qM * (1 - kI); 0];

    BBB4 = [ 1/(1 - kI*kR) * kI/qM * (1 - kR - kR * vdON);
             1/(1 - kI*kR) * kR/qM * (1 - kI + vdON)];

    BBB3a = [ 1/(1 - kI*kR) * kI/qM * (1 - kR + vbON); ...
              1/(1 - kI*kR) * kR/qM * (1 - kI - kI * vbON); 0];

    BBB4a = [ 1/(1 - kI*kR) * kI/qM * (1 - kR + vbON - kR * vdON);
              1/(1 - kI*kR) * kR/qM * (1 - kI + vdON - kI * vbON)];
         
    % direct or reverse coupling?
    if (kI > 0 && kR > 0 && qM > 0)
        % direct coupling
%        [varargout{1:nargout}] = solver4_direct (iinv0, irec0, vka0, theta2, AAA1, AAA2, AAA3, AAA4, BBB1, BBB2, BBB3, BBB4, vdON, vbON);
        [varargout{1:nargout}] = solver4_any ([iinv0; irec0; 0; vka0], theta2, AAA1, AAA2, AAA3, AAA4, AAA3a, AAA4a, BBB1, BBB2, BBB3, BBB4, BBB3a, BBB4a, vdON, vbON);
    elseif (kI < 0 && kR < 0 && qM < 0)
        % reverse coupling
%        [varargout{1:nargout}] = solver4_reverse (iinv0, irec0, vka0, theta2, AAA4, AAA3, AAA2, AAA1, BBB4, BBB3, BBB2, BBB1, vdON, vbON);
        [varargout{1:nargout}] = solver4_any ([iinv0; irec0; 0; vka0], theta2, AAA1, AAA2, AAA3, AAA4, AAA3a, AAA4a, BBB1, BBB2, BBB3, BBB4, BBB3a, BBB4a, vdON, vbON);
    else
        % Errore nei parametri
        [varargout{1:nargout}] = reterror(iinv0, irec0, vka0);
        warning('undefined coupling type.')
        return
    end
end

function varargout = solver4_any (X0, theta2, AA1, AA2, AA3, AA4, AA3a, AA4a, BB1, BB2, BB3, BB4, BB3a, BB4a, vdON, vbON)
    % any coupling type - X0 initial condition (iinv0, irec0, vCinv0, vCrec0) at theta0 = 0 
    % MOS is externally ON from theta=0 to theta=theta2
    
    % Initialization: 
        % chi is the characteristic function of a set
        % fZarray is an array of anonymous function to be composed -- first one is given by the initial conditions
        % iirec is the integral of irec between 0 and 2*pi to be composed zone by zone
        % X0 is the initial values vector, while theta0 the initial normalized time
    chi = @(theta,thetai,thetaf) (thetai<theta)*(theta<thetaf+1e-7); 
    
    fZarray = {@(theta)(theta<=0)*X0};
    thetaarray = [];
    zonestring = [];
    
    iirec = 0;
    theta0 = 0;
%    lastboff = 0;    % lastboff is updated ever time the body diode is off
    firstbon = 2*pi; % firstbon is update onl the first time the bod diode is on
                     % if firstbon == 2*pi the diode has not been turned on, and 
                     % the return value for lastboff = 2*pi; elsewhere it
                     % is the first time the diode has been turned ON
    
    % 10/4/2020
    % new flexible condition on the diodes + cap
    %      |      |      .
    %   -------  \/ ik   |\
    %   |     |           |
    %  ---    |            |
    %  / \   ---           | vka
    %  ---   ---          |
    %   |     |          |
    %   -------
    %      |
    % diode is on if vka = -vON, ik < 0
    % diode is off if vka > -vON, any ik = 0
    % However, due to numerical error, some 'tolerance' must be granted
    vTol = 1e-7;
    iTol = 1e-7;
    diodeisON = @(ik, vka, vON) (ik/iTol <= 1+(vka+vON)*((vka+vON)^2/vTol^2+16)/16/vTol && ik/iTol<=1+(vka+vON)*((vka+vON)^2/vTol^2+16)/16/vTol);
    diodeisOFF = @(ik, vka, vON) (ik/iTol > 1-(vka+vON)*((vka+vON)^2/vTol^2+16)/16/vTol);
    
    
    % Note: the main problem is identifying if (any) diode has been just turned off to select the 
    % correct working zone. If vka = -vdON it is ON, if vka = -vdON + 2.5e-7 it has just been turned off
    % (e.g.: vka > -vdon + 1e-7 the rectifying diode is on)

    % from 0 to theta2: the MOS is ON
    while theta0 < theta2 - 1e-7  % allow tolerance to cope with numerical errors
        % INITIAL CONDITION IDENTIFICATION: at theta0 = 0: 
        % if vCrec(theta0) > -vdON the diode is ON (Z3)
        % if vCrec(theta0) <= -vdON && irec(0) <= 0 the diode is OFF (Z4)
        % if vCrec(theta0) <= vdON && irec(0) >0 there is an error (need to be handled)

        %diodeisOFF(X0(2), X0(4), vdON);
        %diodeisON(X0(2), X0(4), vdON);

        if diodeisOFF(X0(2), X0(4), vdON)
%        if X0(4) > -vdON + 1e-7 % vCrec(theta0) > -vdON + 1e-6
            % Running in zone Z3: MOS is on, diode is off
            %
            % Mathematical model:
            % X'(theta) = AA3 * X(theta) + BB3 (3th order) && vCinv(theta) = 0
            % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCrec(theta)
            %
            % The Z3 may end:
            % - at theta3, when vCrec+vdON turns negative (diode turns on)
            % - at theta2 (MOS externally turns off) 
            % MOS is on, so no need to look for body diode turn on time (vbON -> inf)

            [theta3, final3, iirec3, fZ3] = solver4_Z3(AA3, BB3, X0([1,2,4]), theta0, theta2, vdON, inf);

            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,theta3)*[1,0,0;0,1,0;0,0,0;0,0,1]*fZ3(theta)}];  
            thetaarray = [thetaarray theta3];
            zonestring = [zonestring 'Z_3'];
            
            iirec = iirec + iirec3;
            theta0 = theta3;
%            X0 = [final3(1:2); 2.5e-7; final3(3)]; % body diode off, rectifier diode cannot turn off
            X0 = [final3(1:2); 2.5*vTol; final3(3)]; % body diode off, rectifier diode cannot turn off
            
        elseif diodeisON(X0(2), X0(4), vdON)
%        elseif X0(2) <= 0 % irec(theta0) <= 0
            % Entering in zone Z4: MOS is on, diode is on
            %
            % Mathematical model:
            % X'(theta) = AA4 * X(theta) + BB4 (2th order) && vCinv(theta) = 0 && vCrec(theta) = -vdON
            % X[1] <- iinv(theta), X[2] <- irec(theta)
            %
            % The Z4 may end:
            % - at theta1, when irec(theta) turns positive (diode turns off) 
            % - at theta2, when the MOS is externally turned off
            % no need to look for body diode turn on time (vbON -> inf)

            [theta4, final4, iirec4, fZ4] = solver4_Z4(AA4, BB4, X0(1:2), theta0, theta2, vdON, inf);
            
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,theta4)*[fZ4(theta);0;-vdON]}];
            thetaarray = [thetaarray theta4];
            zonestring = [zonestring 'Z_4'];

            iirec = iirec + iirec4;
            theta0 = theta4;
            if theta4 < theta2
                % here the rectifier diode has just turned off
%                X0 = [final4(1:2); 2.5e-7; -vdON+2.5e-7]; % force body diode to be OFF, rectifier diode to be OFF
                X0 = [final4(1:2); 2.5*vTol; -vdON+2.5*vTol]; % force body diode to be OFF, rectifier diode to be OFF
            else
%                X0 = [final4(1:2); 2.5e-7; -vdON]; % force body diode to be OFF, rectifier diode to be ON
                X0 = [final4(1:2); 2.5*vTol; -vdON]; % force body diode to be OFF, rectifier diode to be ON
            end
        else
            % this is an impossible zone -- to be handled as it can be reached during minimization 
            % growing iirec to inf, and constant voltages and currents up to theta2
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,theta2)*X0}];  
            thetaarray = [thetaarray -1];
            zonestring = [zonestring 'X'];
            
            theta0 = theta2;
            iirec = inf;
        end
    end % end while
    
    % from theta2 to 2*pi: the MOS is off
    while theta0 < 2*pi - 1e-7 % allow tolerance to cope with numerical error
        % INITIAL CONDITION IDENTIFICATION
        % The MOS is off, but the body diode can turn on at any time. 
        % Main hypothesis: the body diode cannot be on at theta2 (dont know how to handle it...)
        % At theta0: 
        % if vCinv(theta0) > -vbON && vCrec(theta0) > -vdON : both diodes are OFF (Z2)
        % if vCinv(theta0) > -vbON && vCrec(theta0) <= -vbON && irec < 0 : body diode off, rectifier diode on (Z1)
        % if vCrec(theta0) <= -vbON && irec(0) <= 0 && vCrec(theta0) > -vdON : body diode on, rectifier diode off (Z3a)
        % if vCrec(theta0) <= -vbON && irec(0) <= 0 && vCrec(theta0) <= -vdON && irec(0) <= 0 : both diode are on (Z4a)
        % otherwise : there is an error (need to be handled)
               
        if diodeisOFF(X0(1), X0(3), vbON) && diodeisOFF(X0(2), X0(4), vdON)
%        if X0(3) > -vbON + 1e-7 && X0(4) > -vdON + 1e-7 % vCrec(theta0) > -vdON + 1e-7, vCinv(theta0) > -vbON + 1e-7
            % Entering in zone Z2: body diode is off, rectifier diode is off
            %
            % Mathematical model:
            % X'(theta) = AA2 * X(theta) + BB2 (4th order)
            % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCinv(theta), X[4] <- vCrec(theta)
            %
            % The Z2 may end:
            % - at theta3, when vCrec+vdON turns negative (rectifier diode turns on)
            % - at theta3, when vCinv+vbON turns negative (body diode turns on) -- only  vbON < inf
            % - at 2*pi, end of period

            [theta2, final2, iirec2, fZ2] = solver4_Z2(AA2, BB2, X0, theta0, 2*pi, vdON, vbON);
            
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,min(theta2))*fZ2(theta)}];
            thetaarray = [thetaarray min(theta2)];
            zonestring = [zonestring 'Z_2'];

            iirec = iirec + iirec2;
            theta0 = min(theta2);
%            lastboff = min(theta2);
            X0 = final2; % no diodes can turn off

        elseif diodeisOFF(X0(1), X0(3), vbON) && diodeisON(X0(2), X0(4), vdON)
%        elseif X0(3) > -vbON + 1e-7 && X0(2) <= 0 % irec(theta0) <= 0
            % Entering in zone Z1: body diode is off, rectifier diode is on
            %
            % Mathematical model:
            % X'(theta) = AA1 * X(theta) + BB1 (3rd order) && vCrec(theta) = -vdON
            % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCinv(theta)
            % X(theta0) = [iinv(theta0); irec0(theta0); 0]
            %
            % The Z1 may end:
            % - at theta3, when irec(theta) turns positive (rectifier diode turns off) 
            % - at theta3, when vCinv(theta)+vbon turns negative (body diode turns on) 
            % - at 2*pi (end of period) 

            [theta1, final1, iirec1, fZ1] = solver4_Z1(AA1, BB1, X0(1:3), theta0, 2*pi, vdON, vbON);
            
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,min(theta1))*[fZ1(theta);-vdON]}];
            thetaarray = [thetaarray min(theta1)];
            zonestring = [zonestring 'Z_1'];

            iirec = iirec + iirec1;
            theta0 = min(theta1);
%            lastboff = min(theta1); 
            % rectifier diode turns off if theta3(1) < theta3(2) && theta3(1) < 2*pi
            % (to be checked only if vbon < inf)
            if size(theta1,2) == 1 || theta1(1) < theta1(2) % no need to check if theta3(1) < 2*pi
%                X0 = [final1(1:3); -vdON+2.5e-7]; % rectifier diode is off
                X0 = [final1(1:3); -vdON+2.5*vTol]; % rectifier diode is off
            else
                X0 = [final1(1:3); -vdON]; % rectifier diode is still on
            end

        elseif diodeisON(X0(1), X0(3), vbON) && diodeisOFF(X0(2), X0(4), vdON)
%        elseif X0(1) <= 0 && X0(4) > -vdON + 1e-7 %
            % Entering in zone Z3(a): body diode is on, rectifier diode is off
            %
            % Mathematical model:
            % X'(theta) = AA3a * X(theta) + BB3a (3th order) && vCinv(theta) = 0
            % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCrec(theta)
            %
            % The Z3(a) may end:
            % - at theta3, when vCrec+vdON turns negative (body diode turns off)
            % - at theta3, when vCrec+vdON turns negative (rectifier diode turns on)
            % - at 2*pi (end of period) 
            % no need to look for body diode turn on time (vbON -> inf)

            % body diode turned ON, recording it
            if firstbon > 2*pi - 1e-7 % unitialized, even considering numerical errors
                firstbon = theta0; end
                
            [theta3, final3, iirec3, fZ3] = solver4_Z3(AA3a, BB3a, X0([1,2,4]), theta0, 2*pi, vdON, vbON);

            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,min(theta3))*([1,0,0;0,1,0;0,0,0;0,0,1]*fZ3(theta)+[0;0;-vbON;0])}];  
            thetaarray = [thetaarray min(theta3)];
            zonestring = [zonestring 'Z_3^a'];

            iirec = iirec + iirec3;
            theta0 = min(theta3);
            % body diode turns off if theta3(2) < theta3(1) && theta3(2) < 2*pi
            % If we are here, then vbon<inf (no need to check)
            if theta3(2) < theta3(1) % no need to check if theta3(2) < 2*pi
%                X0 = [final3(1:2); -vbON+2.5e-7; final3(3)]; % body diode off
                X0 = [final3(1:2); -vbON+2.5*vTol; final3(3)]; % body diode off
            else
                X0 = [final3(1:2); -vbON; final3(3)]; % body diode on
            end

        elseif diodeisON(X0(1), X0(3), vbON) && diodeisON(X0(2), X0(4), vdON)
%        elseif X0(1) <= 0 && X0(2) <= 0 % 
            % Entering in zone Z4(a): body diode is on, rectifier diode is on
            %
            % Mathematical model:
            % X'(theta) = AA4a * X(theta) + BB4a (2th order) && vCinv(theta) = 0 && vCrec(theta) = -vdON
            % X[1] <- iinv(theta), X[2] <- irec(theta)
            %
            % The Z4 may end:
            % - at theta1, when irec(theta) turns positive (rectifier diode turns off) 
            % - at theta1, when iinv(theta) turns positive (body diode turns off) 
            % - at theta2, when the MOS is externally turned off

            % body diode turned ON, recording it
            if firstbon > 2*pi - 1e-7 % unitialized, even considering numerical errors
                firstbon = theta0; end
            
            [theta4, final4, iirec4, fZ4] = solver4_Z4(AA4a, BB4a, X0(1:2), theta0, 2*pi, vdON, vbON);
            
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,min(theta4))*[fZ4(theta);-vbON;-vdON]}];
            thetaarray = [thetaarray min(theta4)];
            zonestring = [zonestring 'Z_4^a'];

            iirec = iirec + iirec4;
            theta0 = min(theta4);
            % which diode has turned on?
            % If we are here, then vbon<inf (no need to check)
            if theta4(1) < theta4(2) % no need to check if theta1(1) < 2*pi
%                X0 = [final4(1:2); -vbON; -vdON+2.5e-7]; % rectifier diode turned off
                X0 = [final4(1:2); -vbON; -vdON+2.5*vTol]; % rectifier diode turned off
            elseif theta4(2) < theta4(1) % no need to check if theta1(2) < 2*pi
%                X0 = [final4(1:2); -vbON+2.5e-7; -vdON]; % body diode turned off
                X0 = [final4(1:2); -vbON+2.5*vTol; -vdON]; % body diode turned off
            elseif theta4(1) < 2*pi % ALMOST IMPOSSIBLE: both diodes off at the same time
%                X0 = [final4(1:2); -vbON+2.5e-7; -vdON+2.5e-7]; % all diodes turned off
                X0 = [final4(1:2); -vbON+2.5*vTol; -vdON+2.5*vTol]; % all diodes turned off
            else
                X0 = [final4(1:2); -vbON; -vdON]; % all diodes still on
            end

        else
            % this is an impossible zone -- to be handled as it can be reached during minimization             
            % growing iirec to inf, and constant voltages and currents up to theta2
            fZarray = [fZarray {@(theta)fZarray{end}(theta)+chi(theta,theta0,theta2)*X0}];  
            thetaarray = [thetaarray -1];
            zonestring = [zonestring 'X'];

            theta0 = 2*pi;
            iirec = inf;
        end

    end % end while

    if (nargout >= 2)
        fZ = @(theta) fZarray{end}(theta);
    end
    
    % Output to be used for minimization
    % v(1:4) is [iinv(2*pi); irec(2*pi); vds(2*pi); vka(2*pi)]
    %        should be equal to [iinv(0); irec(0); vds(0); vka(0)]
    %        with vds(2*pi)=0 for ZVS, iinv(2*pi)=0 for ZVDS
    % v(5) is 'lastboff' the last time when body diode was off
    %        if not 2*pi the body diode id ON when the mos is turned ON
    % v(6) is 'iirec' integrate(irec(theta), 0, 2*pi) to be set to -2*pi
%    v = [X0; lastboff; iirec];
    v = [X0; firstbon; iirec];
    
    switch nargout
        case {0, 1}
            varargout = {v};
        case 2
            varargout = {v, fZ};
        case 3
            varargout = {v, fZ, thetaarray};
        case 4
            varargout = {v, fZ, thetaarray, zonestring};
        otherwise
            warning('Wrong output argument number, no output set.')
    end

end

function varargout = solver4_direct (iinv0, irec0, vCrec0, theta2, AA1, AA2, AA3, AA4, BB1, BB2, BB3, BB4, vdON, ~)
        
    % Direct coupling -- initial values:
    % iinv(0) = iinv0, irec(0) = irec0, vCinv(0) = 0, vCrec(0) = vka0
    % Zone sequence: Z3 -> Z4 -> Z1 -> Z1 (to be expanded on other sequences)

    % TO DO: running in Z3 is assumed to be true, no checkong... 

    % Zone Z3: MOS is on, diode is off
    %
    % X'(theta) = AA3 * X(theta) + BB3 (3th order) && vCinv(theta) = 0
    % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCrec(theta)
    % X(0) = [iinv0; irec0; vCrec0] 
    %
    % The Z3 ends at theta3, when vCrec+vdON turns negative 
    % (diode turns on)

    [theta3, final3, ii3, fZ3] = solver4_Z3(AA3, BB3, [iinv0; irec0; vCrec0], 0, theta2, vdON, inf);

    % Zone Z4: MOS on, diode on
    %
    % X'(theta) = AA4 * X(theta) + BB4 (2th order) && vCinv(theta) = 0 && vCrec(theta) = -vdON
    % X[1] <- iinv(theta), X[2] <- irec(theta)
    % X(theta3) = [iinv(theta3); irec(theta3)] 
    %
    % The Z4 at theta2, when the MOS is externally turned off

    [final4, ii4, fZ4] = solver4_Z4(AA4, BB4, final3(1:2), theta3, theta2, vdON, inf);
    
    % Zone Z1: MOS off, diode on
    %
    % X'(theta) = AA1 * X(theta) + BB1 (3rd order) && vCrec(theta) = -vdON
    % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCinv(theta)
    % X(theta2) = [iinv(theta2); irec0(theta2); 0]
    %
    % The Z1 ends at theta1, quando irec(theta) turns positive 
    % (diode turns off)

  
    [theta1, final1, ii1, fZ1] = solver4_Z1(AA1, BB1, [final4; 0], theta2, 2*pi, vdON, inf);


    % Zone Z2: MOS off, diode off
    %
    % X'(theta) = AA2 * X(theta) + BB2 (4th order)
    % X[1] <- iinv(theta), X[2] <- irec(theta), X[3] <- vCinv(theta), X[4] <- vCrec(theta)
    % X(theta1) = [iinv(theta1); irec(theta1); vCinv(theta1); -vdON] 
    %
    % ends at theta2, MOS externally forced off
    
    [final2, ii2, fZ2] = solver4_Z2(AA2, BB2, [final1; -vdON], theta1, 2*pi, vdON, inf);


    % plot grafici su richiesta
    if (nargout >= 2)


        fZ1full = @(theta) [fZ1(theta); -vdON];
        fZ3full = @(theta) [1 0 0; 0 1 0; 0 0 0; 0 0 1]*fZ3(theta);
        fZ4full = @(theta) [fZ4(theta); 0; -vdON];

        % no other way to compose the fZ
        % yet, this part is not called in optimization, so should not
        % reduce preformance
        %
        fZ = @(theta) (theta <= theta3)*fZ3full(theta) + ...
                      (theta3 < theta)*(theta <= theta2)*fZ4full(theta) + ...
                      (theta2 < theta)*(theta <= theta1)*fZ1full(theta) + ...
                      (theta1 < theta)*fZ2(theta);

    end
    
    v = [final2; ii4+ii3+ii2+ii1];
    
    switch nargout
        case {0, 1}
            varargout = {v};
        case 2
            varargout = {v, fZ};
        case 3
            varargout = {v, fZ, [theta3; theta2; theta1]};
        otherwise
            warning('Wrong output argument number, no output set.')
    end

end
function varargout = solver4_reverse (iinv0, irec0, ~, theta2, AA4, AA3, AA2, AA1, BB4, BB3, BB2, BB1, vdON, ~)
        
    % 180 degrees out of phase coupling - initial values
    % iLinv(0) = iLinv0, iLrec(0) = iLrec0, vCinv(0) = 0, vCrec(0) = -vdON
    % sequence is: Z4 -> Z3 -> Z2 -> Z1

    
    % Zone Z4: MOS is on, diode is on
    %
    % X'(theta) = AA4 * X(theta) + BB4 (2th order) && vCinv(theta) = 0 && vCrec(theta) = -vdON
    % X[1] <- iLinv(theta), X[2] <- iLrec(theta)
    % X(0) = [iLinv0; iLrec0] 
    %
    % The Z4 ends at theta1, when irec(theta) turns positive
    % (diode turns on)

    [theta1, final4, ii4, fZ4] = solver4_Z4(AA4, BB4, [iinv0; irec0], 0, theta2, vdON, inf);

    
    % Zone Z3: MOS still on, diode turned off
    %
    % X'(theta) = AA3 * X(theta) + BB3 (3th order) && vCinv(theta) = 0
    % X[1] <- iLinv(theta), X[2] <- iLrec(theta), X[3] <- vCrec(theta)
    % X(theta1) = [iLinv(theta1); iLrec(theta1); vCrec(theta1)] 
    %
    % The Z3 ends at theta2, when the MOS is externally turned off

    [final3, ii3, fZ3] = solver4_Z3(AA3, BB3, [final4; -vdON], theta1, theta2, vdON, inf);

    
    % Zone Z2: MOS off, diode off
    %
    % X'(theta) = AA2 * X(theta) + BB2 (4th order)
    % X[1] <- iLinv(theta), X[2] <- iLrec(theta), X[3] <- vCinv(theta), X[4] <- vCrec(theta)
    % X(theta2) = [iLinv(theta2); iLrec(theta2); 0; vCrec(theta2)] 
    %
    % ends at theta3, diode tunrs itself on
    
    [theta3, final2, ii2, fZ2] = solver4_Z2(AA2, BB2, [final3(1:2); 0; final3(3)], theta2, 2*pi, vdON, inf);


    % Zone Z1: MOS still off, diode turs ifself on
    %
    % X'(theta) = AA1 * X(theta) + BB1 (3rd order) && vCrec(theta) = -vdON
    % X[1] <- iLinv(theta), X[2] <- iLrec(theta), X[3] <- vCinv(theta)
    % X(0) = [iLinv(theta2); iLrec(theta2); vCinv(theta3)]
    %
    % La zona 1 finisce al tempo 2*pi, MOS acceso esternamento

    [final1, ii1, fZ1] = solver4_Z1(AA1, BB1, final2(1:3), theta3, 2*pi, vdON, inf);

    
    % plot grafici su richiesta
    if (nargout >= 2)


        fZ4full = @(theta) [fZ4(theta); 0; -vdON];
        fZ3full = @(theta) [1 0 0; 0 1 0; 0 0 0; 0 0 1]*fZ3(theta);
        fZ1full = @(theta) [fZ1(theta); -vdON];

        % non ho trovato altro modo di fare la composizione
        % Ma non dovrebbe esserci problemi di velocit�
        % questo pezzo chiamato solo raramente
        %
        fZ = @(theta) (theta <= theta1)*fZ4full(theta) + ...
                      (theta1 < theta)*(theta <= theta2)*fZ3full(theta) + ...
                      (theta2 < theta)*(theta <= theta3)*fZ2(theta) + ...
                      (theta3 < theta)*fZ1full(theta);

    end
    
    v = [final1; -vdON; ii1+ii2+ii3+ii4];
    
    switch nargout
        case {0, 1}
            varargout = {v};
        case 2
            varargout = {v, fZ};
        case 3
            varargout = {v, fZ, [theta1; theta2; theta3]};
        otherwise
            warning('Wrong output argument number, no output set.')
    end

end

% THESE FUNCTION MAY BE REWRITTEN:
% - double output for the two zero checking
% - no more need for zero checking flag
% - dont know how to express the angles

%function [theta1, final1, ii1, fZ1] = solvernew1_Z1(AA1, BB1, initial1, theta0, theta1sup, vdON)
function varargout = solver4_Z1(AA1, BB1, initial1, thetai, thetaf, ~, vbON)
    % ZONE Z1
    %
    % The MOS is OFF and the diode is ON
    % Zone ends when either:
    % - MOS is turned ON (externally), or
    % - body diode turns ON (vCinv + vbON < 0)
    % - rectifying diode turns OFF (irec > 0)

    % No significative advantage in declaring FF13 & iFF13 as global variables
    FF13 = @(theta, sigma0, omega0, lambda3) ...
        [ exp(theta * sigma0) * cos(theta * omega0), exp(theta * sigma0) * sin(theta * omega0), 0; ...
          -exp(theta * sigma0) * sin(theta * omega0), exp(theta * sigma0) * cos(theta * omega0), 0; ...
          0, 0, exp(theta * lambda3) ];
    iFF13 = @(thetai, thetaf, sigma0, omega0, lambda3) ...
        [ (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
          (exp(thetaf * sigma0) * (-omega0 * cos(thetaf * omega0) + sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (omega0 * cos(thetai * omega0) - sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2),  0; ...
          (exp(thetaf * sigma0) * (omega0 * cos(thetaf * omega0) - sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (-omega0 * cos(thetai * omega0) + sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
          (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), 0 ; ...
          0, 0, (exp(thetaf * lambda3) - exp(thetai * lambda3))/lambda3];
%    Z1flag = false; % used to generate an error if fzero fails

    % Questa � la soluzione particolare, cambia a seconda se ci sono colonne nulle in AA1
    % aux1 � il termine costante, lin1 il termine lineare in theta della soluzione particolare [ aux1 + lin1 * theta ] 
    if (any(AA1(:,2)))
        %aux1 = -inv(AA1)*BB1; % generates warning
        aux1 = -AA1\BB1;
        lin1 = [0; 0; 0];
    else
        %aux1 = -[1, 0; 0, 0; 0, 1]*inv(AA1([1,3],[1,3]))*BB1([1,3]); % termine costante qaundo seconda colonna di AA1 � nulla
        aux1 = -[1, 0; 0, 0; 0, 1]*(AA1([1,3],[1,3])\BB1([1,3])); % termine costante quando seconda colonna di AA1 � nulla
        lin1 = [0; BB1(2) - [AA1(2,1), AA1(2,3)]*(AA1([1,3],[1,3])\BB1([1,3])); 0]; % termine lineare in theta quando seconda colonna di AA1 � nulla
        iFF13 = @(thetai, thetaf, sigma0, omega0, lambda3) ...
            [ (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
              (exp(thetaf * sigma0) * (-omega0 * cos(thetaf * omega0) + sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (omega0 * cos(thetai * omega0) - sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2),  0; ...
              (exp(thetaf * sigma0) * (omega0 * cos(thetaf * omega0) - sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (-omega0 * cos(thetai * omega0) + sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
              (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), 0 ; ...
              0, 0, thetaf-thetai];
    end

    
    [VV1, DD1] = eig(AA1); LL1 = diag(DD1); % VV1 <- autovettori; ll1 <- autovalori
    % riordine autovalori: primi due complessi coniugati, terzo reale negativo
    %VV1 = [VV1(:,find(imag(LL1)>0)), VV1(:,find(imag(LL1)<0)), VV1(:,find(imag(LL1)==0))];
    VV1 = [VV1(:,imag(LL1)>0), VV1(:,imag(LL1)<0), VV1(:,imag(LL1)==0)];
    %LL1 = [LL1(find(imag(LL1)>0)); LL1(find(imag(LL1)<0)); LL1(find(imag(LL1)==0))];
    LL1 = [LL1(imag(LL1)>0); LL1(imag(LL1)<0); LL1(imag(LL1)==0)];

    WW1 = [real(VV1(:,1)), imag(VV1(:,1)), VV1(:,3)]; % Questo deve essere reale
    %CC1 = FF13(-thetai, real(LL1(1)), imag(LL1(1)), LL1(3))*inv(WW1)*(initial1 - aux1 - lin1*thetai);
    CC1 = FF13(-thetai, real(LL1(1)), imag(LL1(1)), LL1(3))*(WW1\(initial1 - aux1 - lin1*thetai));

    if (nargout == 4)
        % se chiedo 3 argomenti [final1, ii1, fZ1]: sto forzando io la fine della zona a thetaf 
        % altrimenti, chiedo 4 argomenti [thetaf, final1, ii1, fZ1]: sto guardando quando forzando io la fine della zona a thetaf 
        % Devo trovare quando iLrec(theta) diventa positiva - il 5o argomento (thetaf) � da intendere come il sup
        % La sola funzione di iLrec � data da
        irecZ = @(theta) WW1(2,:)*FF13(theta, real(LL1(1)), imag(LL1(1)), LL1(3))*CC1 + aux1(2) + lin1(2)*theta;
        [thetaf1, ~] = myfzero(irecZ, [thetai, thetaf]);
        
        % look for body diode turn on time (only vbON < inf)
        % vCinv(theta)+vbON turns negative - result between thetai and thetaf
        if vbON < inf
            vCinvZ = @(theta) WW1(3,:)*FF13(theta, real(LL1(1)), imag(LL1(1)), LL1(3))*CC1 + aux1(3) + lin1(3)*theta + vbON;
            [thetaf2, ~] = myfzero(vCinvZ, [thetai, thetaf]);
            thetaf = min(thetaf1, thetaf2);
        else
            thetaf = thetaf1;
        end
    end

%    if Z1flag
%        ii1 = inf;
%    else
        % contributo all'integrale di iLrec(theta) in (0, 2*pi)
        ii1 = WW1(2,:)*iFF13(thetai, thetaf, real(LL1(1)), imag(LL1(1)), LL1(3))*CC1+aux1(2)*(thetaf-thetai)+lin1(2)*(thetaf^2-thetai^2)/2;
%    end

%    if showplot
        % build the drawing function only if necessary
        fZ1 = @(theta) WW1*FF13(theta, real(LL1(1)), imag(LL1(1)), LL1(3))*CC1 + aux1 + lin1*theta;
        final1 = fZ1(thetaf);
%    else
%        % otherwise, simply compute the final value (@ thetaf)
%        fZ1 = [];
%        final1 = WW1*FF13(thetaf, real(LL1(1)), imag(LL1(1)), LL1(3))*CC1 + aux1 + lin1*thetaf;
%    end
        
    if (nargout == 3)
        % se chiedo 3 argomenti [final1, ii1, fZ1]: sto forzando io la fine della zona a thetaf 
        varargout = {final1, ii1, fZ1}; 
    else
        % altrimenti, chiedo 4 argomenti [thetaf, final1, ii1, fZ1]: sto guardando quando forzando io la fine della zona a thetaf 
        if vbON < inf
            varargout = {[thetaf1, thetaf2], final1, ii1, fZ1}; 
        else
            varargout = {thetaf, final1, ii1, fZ1}; 
        end
    end

end


%function [final2, ii2, fZ2] = solvernew1_Z2(AA2, BB2, initial2, thetai, thetaf, vdON)
function varargout = solver4_Z2(AA2, BB2, initial2, thetai, thetaf, vdON, vbON)
    % ZONE Z2
    %
    % The MOS is OFF and the diode is OFF
    % Zone ends when either:
    % - MOS is turned ON (externally), or
    % - body diode turns ON (vCinv + vbON < 0)
    % - rectifying diode turns ON (vCrec + vdON < 0)

    % No significative advantage in declaring FF2 & iFF2 as global variables
    FF2 = @(theta, sigma1, omega1, sigma2, omega2) ...
        [ exp(theta * sigma1) * cos(theta * omega1), exp(theta * sigma1) * sin(theta * omega1), 0, 0; ...
          -exp(theta * sigma1) * sin(theta * omega1), exp(theta * sigma1) * cos(theta * omega1), 0, 0; ...
          0, 0, exp(theta * sigma2) * cos(theta * omega2), exp(theta * sigma2) * sin(theta * omega2);
          0, 0, -exp(theta * sigma2) * sin(theta * omega2), exp(theta * sigma2) * cos(theta * omega2)];
    iFF2 = @(thetai, thetaf, sigma1, omega1, sigma2, omega2) ...
        [ (exp(thetaf * sigma1) * (sigma1 * cos(thetaf * omega1) + omega1 * sin(thetaf * omega1)) - exp(thetai * sigma1) * (sigma1 * cos(thetai * omega1) + omega1 * sin(thetai * omega1)))/(sigma1^2 + omega1^2), ...
          (exp(thetaf * sigma1) * (-omega1 * cos(thetaf * omega1) + sigma1 * sin(thetaf * omega1)) + exp(thetai * sigma1) * (omega1 * cos(thetai * omega1) - sigma1 * sin(thetai * omega1)))/(sigma1^2 + omega1^2), 0, 0; ...
          (exp(thetaf * sigma1) * (omega1 * cos(thetaf * omega1) - sigma1 * sin(thetaf * omega1)) + exp(thetai * sigma1) * (-omega1 * cos(thetai * omega1) + sigma1 * sin(thetai * omega1)))/(sigma1^2 + omega1^2), ...
          (exp(thetaf * sigma1) * (sigma1 * cos(thetaf * omega1) + omega1 * sin(thetaf * omega1)) - exp(thetai * sigma1) * (sigma1 * cos(thetai * omega1) + omega1 * sin(thetai * omega1)))/(sigma1 ^2 + omega1^2), 0, 0; ...
          0, 0, (exp(thetaf * sigma2) * (sigma2 * cos(thetaf * omega2) + omega2 *sin(thetaf * omega2)) - exp(thetai * sigma2) * (sigma2 * cos(thetai * omega2) + omega2 * sin(thetai * omega2)))/(sigma2^2 + omega2^2), ...
          (exp(thetaf * sigma2) * (-omega2 * cos(thetaf * omega2) + sigma2 *sin(thetaf * omega2)) + exp(thetai * sigma2) * (omega2 * cos(thetai * omega2) - sigma2 * sin(thetai * omega2)))/(sigma2^2 + omega2^2); ...
          0, 0, (exp(thetaf * sigma2) * (omega2 * cos(thetaf * omega2) - sigma2 *sin(thetaf * omega2)) + exp(thetai * sigma2) * (-omega2 * cos(thetai * omega2) + sigma2 * sin(thetai * omega2)))/(sigma2^2 + omega2^2), ...
          (exp(thetaf * sigma2) * (sigma2 * cos(thetaf * omega2) + omega2 *sin(thetaf * omega2)) - exp(thetai * sigma2) * (sigma2 * cos(thetai * omega2) + omega2 * sin(thetai * omega2)))/(sigma2^2 + omega2^2)];
%    Z2flag = false;

    %aux2 = inv(AA2)*BB2; % this generates warning
    aux2 = AA2\BB2;
    [VV2, DD2] = eig(AA2); LL2 = diag(DD2); % VV2 <- autovettori; DD2 <- autovalori - a due a due complessi coniugati
    % non dovrebbe esserci bisogno di riordinaee i vettori

    WW2 = [real(VV2(:,1)), imag(VV2(:,1)), real(VV2(:,3)), imag(VV2(:,3))]; % Questo deve essere reale
    %CC2 = FF2(-theta1, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*inv(WW2)*(aux2 + [initial2; -vdON]);
    CC2 = FF2(-thetai, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*(WW2\(aux2 + initial2));

    if (nargout == 4)
        % se chiedo 3 argomenti [final2, ii2, fZ2]: sto forzando io la fine della zona a thetaf 
        % altrimenti, chiedo 4 argomenti [thetaf, final2, ii2, fZ2]: sto guardando quando forzando io la fine della zona a thetaf 
        % Devo trovare quando vCrec(theta)+vdON diventa negativa - il 5o argomento (thetaf) � da intendere come il sup
        % La sola funzione di vCrec+vdON � data da
        vCrecZ = @(theta) WW2(4,:)*FF2(theta, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*CC2 - aux2(4) + vdON;
        [thetaf1, ~] = myfzero(vCrecZ, [thetai, thetaf]);
        
        % look for body diode turn on time (only vbON < inf)
        % vCinv(theta)+vbON turns negative - result between thetai and thetaf
        if vbON < inf
            vCinvZ = @(theta) WW2(3,:)*FF2(theta, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*CC2 - aux2(3) + vbON;
            [thetaf2, ~] = myfzero(vCinvZ, [thetai, thetaf]);
            thetaf = min(thetaf1, thetaf2);
        else
            thetaf = thetaf1;
        end

    end

%    if Z2flag
%        ii2 = inf;
%    else
        % contributo all'integrale di iLrec(theta) in (0, 2*pi)
        iirec2 = WW2(2,:)*iFF2(thetai, thetaf, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*CC2-aux2(2)*(thetaf-thetai);
%    end

%    if showplot
        % build the drawing function only if necessary
        % le quattro variabili sono date da (non necessario tutto)
        fZ2 = @(theta) WW2*FF2(theta, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*CC2 - aux2;
        final2 = fZ2(thetaf);
%    else
%        % otherwise, simply compute the final value (@ thetaf)
%        fZ2 = [];
%        final2 = WW2*FF2(thetaf, real(LL2(1)), imag(LL2(1)), real(LL2(3)), imag(LL2(3)))*CC2 - aux2;
%    end
    
    if (nargout == 3)
        % se chiedo 3 argomenti [final2, ii2, fZ2]: sto forzando io la fine della zona a thetaf 
        varargout = {final2, iirec2, fZ2}; 
    else
        % altrimenti, chiedo 4 argomenti [thetaf, final2, ii2, fZ2]: sto guardando quando forzando io la fine della zona a thetaf 
        if vbON < inf
            varargout = {[thetaf1, thetaf2], final2, iirec2, fZ2}; 
        else
            varargout = {thetaf, final2, iirec2, fZ2}; 
        end
    end
    
end

%function [theta3, final3, ii3, fZ3] = solvernew1_Z3(AA3, BB3, initial3, theta2, theta3sup, vdON)
function varargout = solver4_Z3(AA3, BB3, initial3, thetai, thetaf, vdON, vbON)
    % ZONE Z3
    %
    % The MOS/body diode is ON (vCinv = 0 / vCinv = -vbON) and the rectifier diode is OFF
    % Zone ends when either:
    % - the MOS is turned OFF (externally), or
    % - the body diode turns off (iinv > 0) -- only vbON < inf 
    % - the rectifier diode turns on (vCrec + vdON < 0)

    % No significative advantage in declaring FF13 & iFF13 as global variables
    FF13 = @(theta, sigma0, omega0, lambda3) ...
        [ exp(theta * sigma0) * cos(theta * omega0), exp(theta * sigma0) * sin(theta * omega0), 0; ...
          -exp(theta * sigma0) * sin(theta * omega0), exp(theta * sigma0) * cos(theta * omega0), 0; ...
          0, 0, exp(theta * lambda3) ];
    iFF13 = @(thetai, thetaf, sigma0, omega0, lambda3) ...
        [ (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
          (exp(thetaf * sigma0) * (-omega0 * cos(thetaf * omega0) + sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (omega0 * cos(thetai * omega0) - sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2),  0; ...
          (exp(thetaf * sigma0) * (omega0 * cos(thetaf * omega0) - sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (-omega0 * cos(thetai * omega0) + sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
          (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), 0 ; ...
          0, 0, (exp(thetaf * lambda3) - exp(thetai * lambda3))/lambda3];
%    Z3flag = false;

    % Questa � la soluzione particolaria, cambia a seconda se ci sono colonne nulle in AA3
    % aux3 � il termine costante, lin3 il termine lineare in theta della soluzione particolare [ aux3 + lin3 * theta ] 
    if (any(AA3(:,1)))
        %aux1 = -inv(AA3)*BB3; % generates warning
        aux3 = -AA3\BB3;
        lin3 = [0; 0; 0];
    else
        %aux3 = -[1, 0; 0, 0; 0, 1]*inv(AA3([2,3],[2,3]))*BB3([2,3]); % termine costante qaundo prima colonna di AA3 � nulla
        aux3 = -[0, 0; 1, 0; 0, 1]*(AA3([2,3],[2,3])\BB3([2,3])); % termine costante quando prima colonna di AA3 � nulla
        lin3 = [BB3(1) - [AA3(1,2), AA3(1,3)]*(AA3([2,3],[2,3])\BB3([2,3])); 0; 0]; % termine lineare in theta quando prima colonna di AA3 � nulla
        iFF13 = @(thetai, thetaf, sigma0, omega0, lambda3) ...
            [ (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
              (exp(thetaf * sigma0) * (-omega0 * cos(thetaf * omega0) + sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (omega0 * cos(thetai * omega0) - sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2),  0; ...
              (exp(thetaf * sigma0) * (omega0 * cos(thetaf * omega0) - sigma0 *sin(thetaf * omega0)) + exp(thetai * sigma0) * (-omega0 * cos(thetai * omega0) + sigma0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), ...
              (exp(thetaf * sigma0) * (sigma0 * cos(thetaf * omega0) + omega0 *sin(thetaf * omega0)) - exp(thetai * sigma0) * (sigma0 * cos(thetai * omega0) + omega0 * sin(thetai * omega0)))/(sigma0^2 + omega0^2), 0 ; ...
              0, 0, thetaf-thetai];
    end
     
    %aux3 = inv(AA3)*BB3; % generates warning
    %aux3 = AA3\BB3;
    [VV3, DD3] = eig(AA3); LL3 = diag(DD3); % VV3 <- autovettori; LL3 <- autovalori
    % riordine autovalori: primi due complessi coniugati, terzo reale negativo
    %VV3 = [VV3(:,find(imag(LL3)>0)), VV3(:,find(imag(LL3)<0)), VV3(:,find(imag(LL3)==0))];
    VV3 = [VV3(:,imag(LL3)>0), VV3(:,imag(LL3)<0), VV3(:,imag(LL3)==0)];
    %LL3 = [LL3(find(imag(LL3)>0)); LL3(find(imag(LL3)<0)); LL3(find(imag(LL3)==0))];
    LL3 = [LL3(imag(LL3)>0); LL3(imag(LL3)<0); LL3(imag(LL3)==0)];

    WW3 = [real(VV3(:,1)), imag(VV3(:,1)), VV3(:,3)]; % Questo deve essere reale
    %CC3 = FF13(-theta2, real(LL3(1)), imag(LL3(1)), LL3(3))*inv(WW3)*(aux3 + initial3);
    CC3 = FF13(-thetai, real(LL3(1)), imag(LL3(1)), LL3(3))*(WW3\(initial3 - aux3 - lin3*thetai));

    if (nargout == 4)
        % se chiedo 3 argomenti [final3, ii3, fZ3]: sto forzando io la fine della zona a thetaf 
        % altrimenti, chiedo 4 argomenti [thetaf, final3, ii3, fZ3]: sto guardando quando forzando io la fine della zona a thetaf 
        % Devo trovare quando vCrec(theta)+vdON diventa negativa - il 5o argomento (thetaf) � da intendere come il sup
        % La sola funzione di vCrec+vdON � data da
        vCrecZ = @(theta) WW3(3,:)*FF13(theta, real(LL3(1)), imag(LL3(1)), LL3(3))*CC3 + aux3(3) +lin3(3)*theta + vdON;
        [thetaf1, ~] = myfzero(vCrecZ, [thetai, thetaf]);

        % look for body diode turn off time (only vbON < inf)
        % iinv(theta) turns positive - result between thetai and thetaf
        if vbON < inf
            iinvZ = @(theta) WW3(1,:)*FF13(theta, real(LL3(1)), imag(LL3(1)), LL3(3))*CC3 + aux3(1) +lin3(1)*theta;
            [thetaf2, ~] = myfzero(iinvZ, [thetai, thetaf]);
            thetaf = min(thetaf1, thetaf2);
        else
            thetaf = thetaf1;
        end

    end

%    if Z3flag
%        ii3 = inf;
%    else
        % contributo all'integrale di irec(theta) in (0, 2*pi)
        iirec3 = WW3(2,:)*iFF13(thetai, thetaf, real(LL3(1)), imag(LL3(1)), LL3(3))*CC3+aux3(2)*(thetaf-thetai)-lin3(2)*(thetaf^2-thetai^2)/2;
%    end

%    if showplot
        % build the drawing function only if necessary
        fZ3 = @(theta) WW3*FF13(theta, real(LL3(1)), imag(LL3(1)), LL3(3))*CC3 + aux3 + lin3*theta;
        final3 = fZ3(thetaf);
%    else
%        % otherwise, simply compute the final value (@ thetaf)
%        fZ3 = [];
%        final3 = WW3*FF13(thetaf, real(LL3(1)), imag(LL3(1)), LL3(3))*CC3 + aux3 +lin3*thetaf;
%    end
    
    if (nargout == 3)
        % se chiedo 3 argomenti [final3, ii3, fZ3]: sto forzando io la fine della zona a thetaf 
        varargout = {final3, iirec3, fZ3}; 
    else
        % altrimenti, chiedo 4 argomenti [thetaf, final3, ii3, fZ3]: sto guardando quando forzando io la fine della zona a thetaf 
        if vbON < inf
            varargout = {[thetaf1, thetaf2], final3, iirec3, fZ3}; 
        else
            varargout = {thetaf, final3, iirec3, fZ3}; 
        end
    end

end

function varargout = solver4_Z4(AA4, BB4, initial4, thetai, thetaf, ~, vbON)
    % ZONE Z4
    %
    % The MOS/body diode is ON (vCinv = 0 or vCinv = -vbON) and the rectifier diode is ON (vCrec = -vdON)
    % Zone ends when either:
    % - the MOS is turned OFF (externally), or
    % - the body diode turns off (iinv > 0) -- only vbON < inf 
    % - the rectifier diode turns off (irec > 0)

    % No significative advantage in declaring FF4 & iFF4 as global variables
    FF4 = @(theta, lambda1, lambda2) ...
        [ exp(theta * lambda1), 0; 0 , exp(theta * lambda2) ];
    iFF4 = @(thetai, thetaf, lambda1, lambda2) ...
        [ (exp(thetaf * lambda1) - exp(thetai * lambda1))/lambda1, 0; ...
          0, (exp(thetaf * lambda2) - exp(thetai * lambda2))/lambda2 ];
%    Z4flag = false;
     
    % Questa � la soluzione particolare, cambia a seconda se ci sono colonne nulle in AA1
    % aux4 � il termine costante, lin4 il termine lineare in theta della soluzione particolare [ aux4 + lin4 * theta ] 
    if (isequal(any(AA4),[1,1])) % due colonne non nulle
        %aux4 = inv(AA4)*BB4; % generates warning
        aux4 = -AA4\BB4; % AA4 invertibile, non c'� termine lineare
        lin4 = [0; 0];
    elseif (isequal(any(AA4),[0,1])) % prima colonna nulla
        aux4 = [0; -BB4(2)/AA4(2,2)];
        lin4= [[1, -AA4(1,2)/AA4(2,2)]*BB4; 0];
        iFF4 = @(thetai, thetaf, lambda1, lambda2) ...
            [ thetaf - thetai, 0; ...
              0, (exp(thetaf * lambda2) - exp(thetai * lambda2))/lambda2 ];    
    elseif (isequal(any(AA4),[1,0])) % seconda colonna nulla
        aux4 = [-BB4(1)/AA4(1,1); 0];
        lin4 = [0; [-AA4(2,1)/AA4(1,1), 1]*BB4];
        iFF4 = @(thetai, thetaf, lambda1, lambda2) ...
            [ thetaf - thetai, 0; ...
              0, (exp(thetaf * lambda2) - exp(thetai * lambda2))/lambda2 ];    
    else % AA4 matrice nulla
        aux4 = [0; 0];
        lin4 = BB4;
        iFF4 = @(thetai, thetaf, lambda1, lambda2) ...
            [ thetaf - thetai, 0; ...
              0, thetaf - thetai ];    
    end
    
    [WW4, DD4] = eig(AA4); LL4 = diag(DD4); % VV4 <- autovettori; LL4 <- autovalori
    % no need to sort eigenvalues, or to convert eigenvectors to real

    %CC4 = FF4(-thetai, LL4(1), LL4(2))*inv(WW4)*(initial4 - aux4);
    CC4 = FF4(-thetai, LL4(1), LL4(2))*(WW4\(initial4 - aux4 - lin4*thetai));

    if (nargout == 4)
        % se chiedo 3 argomenti [final4, ii4, fZ4]: sto forzando io la fine della zona a thetaf 
        % altrimenti, chiedo 4 argomenti [thetaf, final4, ii4, fZ4]: sto guardando quando forzando io la fine della zona a thetaf 
        % Devo trovare quando iLrec(theta) diventa positiva - il 5o argomento (thetaf) � da intendere come il sup
        % La sola funzione di iLrec � data da
        irecZ = @(theta) WW4(2,:)*FF4(theta, LL4(1), LL4(2))*CC4 + aux4(2) + lin4(2)*theta;
        [thetaf1, ~] = myfzero(irecZ, [thetai, thetaf]);
        
        % look for body diode turn off time (only vbON < inf)
        % iinv(theta) turns positive - result between thetai and thetaf
        if vbON < inf
            iinvZ = @(theta) WW4(1,:)*FF4(theta, LL4(1), LL4(2))*CC4 + aux4(1) + lin4(1)*theta;
            [thetaf2, ~] = myfzero(iinvZ, [thetai, thetaf]);
            thetaf = min(thetaf1, thetaf2);
        else
            thetaf = thetaf1;
        end


    end

%    if Z4flag
%        ii4 = inf;
%    else
        % contributo all'integrale di iLrec(theta) in (0, 2*pi)
        iirec4 = WW4(2,:)*iFF4(thetai, thetaf, LL4(1), LL4(2))*CC4+aux4(2)*(thetaf-thetai)+lin4(2)*(thetaf^2-thetai^2)/2;
%    end
    
%    if showplot
        % build the drawing function only if necessary
        fZ4 = @(theta) WW4*FF4(theta, LL4(1), LL4(2))*CC4 + aux4 + lin4 * theta;
        final4 = fZ4(thetaf);
%    else
%        % otherwise, simply compute the final value (@ thetaf)
%        fZ4 = [];
%        final4 = WW4*FF4(thetaf, LL4(1), LL4(2))*CC4 + aux4 + lin4*thetaf;
%    end
        
    if (nargout == 3)
        % se chiedo 3 argomenti [final4, iirec4, fZ4]: sto forzando io la fine della zona a thetaf 
        varargout = {final4, iirec4, fZ4}; 
    else
        % altrimenti, chiedo 4 argomenti [thetaf, final4, iirec4, fZ4]: sto guardando quando forzando io la fine della zona a thetaf 
        if vbON < inf
            varargout = {[thetaf1, thetaf2], final4, iirec4, fZ4}; 
        else
            varargout = {thetaf, final4, iirec4, fZ4}; 
        end
    end
    
end

function [x, flag] = myfzero (myf, x0)

    %x = fzero(f, x0);
    
    % need to find a zero of myf(x):
    % 1. we nee the FIRST solution 
    % 2. the solution has to be in the interval [x0(1) x0(2)]
% 4/11/2019
    % 3. if the solution is almost equal to x0(2), then force it to be
    %    equal to x0(2), since, probably, it is due to numerical errors 


    flag = false;
    
    grid = linspace(x0(1), x0(2), 24);
    values = arrayfun (myf, grid);
    values = values * sign(values(1)); % from positive to negative
    %figure; plot(grid, values);
    
    pp = find(values<0, 1);
    
    if ~isempty(pp)
        % the solution exist; this is the first one
        x = fzero (myf, grid(pp-1:pp), optimset('Display','off'));
        % 4/11/2019
        if x > x0(2) - 1e-4*2*pi % 4/11/2019
            x = x0(2); end % 4/11/2019
        if (x - x0(1)) > 1e-7
            % everything is OK, can return
            return
        end
    end

%    % not able to find a solution... try a better refinement
%    
%    % grid = x0(1):pi/384:(x0(2)+pi/384); % add a point to be sure that the last point > x0(2)
%    grid = linspace(x0(1), x0(2), 384);
%    values = arrayfun (myf, grid);
%    values = values * sign(values(1)); % from positive to negative
%    
%    pp = find(values<0, 1);
%    
%    if ~isempty(pp)
%        % the solution exists; this is the first one
%        x = fzero (myf, grid(pp-1:pp), optimset('Display','off'));
%        % everything is OK, can return
%        return
%    end
%    
%    % Qui va molto male, tutto positivo
%    % mi preparo a non avere soluzioni
%    % This may be commented to speed up in finding solutions
%     try
%         x = fzero(myf, x0(1)/2+x0(2),optimset('Display','off'));
%          % here a solution is found, check if in the limit 
%          if x0(1) > x || x > x0(2) 
%              % out of range! using x0(2)
%              flag = true;
%          end
%     catch
%         % something failed - returning the limit of the allowed interval
%         x = x0(2);
%         flag = true;
%     end
    % no solution exist -- returning the last possible value
    x = x0(2);
    flag = true;
end

function varargout = reterror(iinv0, irec0, vka0)
    switch nargout
        case {0, 1}
            varargout = {[iinv0; irec0; 0; vka0; 0; 0]}; % generates an error
        case 2
            varargout = {[iinv0; irec0; 0; vka0; 0; 0], @(theta) [iinv0; irec0; 0; vka0]};
        case 3
            varargout = {[iinv0; irec0; 0; vka0; 0; 0], @(theta) [iinv0; irec0; 0; vka0], [0; 2*pi]};
        case 4
            varargout = {[iinv0; irec0; 0; vka0; 0; 0], @(theta) [iinv0; irec0; 0; vka0], [0; 2*pi], 'X'};
        otherwise
            warning('Wrong output argument number, no output set.')
    end
end